from pm4py.algo.simulation.playout.petri_net import algorithm as simulator
from tqdm import tqdm


def simulate_model(net, im, fm, repetitions=200, traces=5000):
    logs = []
    for i in tqdm(range(repetitions), desc='Simulating model'):
        log = simulator.apply(net, im, variant=simulator.Variants.BASIC_PLAYOUT, parameters={simulator.Variants.BASIC_PLAYOUT.value.Parameters.NO_TRACES: traces})
        #ensure no zero-length traces are in final log
        for j in range(len(log)):
            tr = log[j]
            if len(tr)==0:
                length = 0
                while length == 0:
                    tr = simulator.apply(net, im, variant=simulator.Variants.BASIC_PLAYOUT, parameters={simulator.Variants.BASIC_PLAYOUT.value.Parameters.NO_TRACES: 1})[0]
                    length = len(tr)
                log[j] = tr
        logs.append(log)
    return logs



import copy
import math
import random
import statistics

from numpy.random import choice

from src.estimation.metrics import estimate_species_richness_chao, get_singletons, get_doubletons, estimate_species_richness_chao_corrected, get_number_observed_species, entropy_exp, \
    simpson_diversity, estimate_shannon_entropy_abundance, estimate_simpson_diversity_abundance, \
    estimate_shannon_entropy_incidence, estimate_simpson_diversity_incidence


def get_bootstrap_probabilities_abundance(reference_sample, sample_size):
    s_obs = len(reference_sample)

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)
    f_0 = 0
    if f_2 > 0:
        f_0 = ((sample_size - 1) / sample_size) * f_1 ** 2 / (2 * f_2)
    else:
        f_0 = ((sample_size - 1) / sample_size) * f_1 * (f_1 - 1) / 2
    f_0 = math.ceil(f_0)
    # print(s_obs,f_0)
    probabilities = []
    c = get_c_n_abundance(reference_sample, sample_size)
    factor = factor_abundance(reference_sample, sample_size, c)

    for x_i in reference_sample.values():
        adapted_p = (x_i / sample_size) * (1 - factor * ((1 - (x_i / sample_size)) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((1 - c) / f_0) for i in range(0, f_0)]
    return probabilities

def get_bootstrap_probabilities_incidence(reference_sample, sample_size):
    s_obs = len(reference_sample)
    f_0 = 0

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2 > 0:
        f_0 = ((sample_size - 1) / sample_size) * f_1 ** 2 / (2 * f_2)
    else:
        f_0 = ((sample_size - 1) / sample_size) * f_1 * (f_1 - 1) / 2
    f_0 = math.ceil(f_0)
    #print(f_0, s_obs)
    # print(s_obs,f_0)

    probabilities = []
    c = get_c_n_incidence(reference_sample, sample_size)
    u = sum(reference_sample.values())

    factor = factor_incidence(reference_sample, sample_size, u, c)

    for y_i in reference_sample.values():
        adapted_p = (y_i / sample_size) * (1 - factor * ((1 - (y_i / sample_size)) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((u / sample_size) * (1 - c) / f_0) for i in range(0, f_0)]
    return probabilities

def generate_sample_abundance(reference_sample, sample_size, probabilities):
    s_obs = len(reference_sample)
    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)
    f_0 = 0
    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0 = math.ceil(f_0)
    selected_species = choice([x for x in range(0, s_obs + f_0)], sample_size, p=probabilities)
    bootstrap_sample = {}
    for s in selected_species:
        bootstrap_sample[s] = bootstrap_sample.get(s, 0) + 1

    return bootstrap_sample

def generate_sample_incidence(reference_sample, sample_size, probabilities):
    s_obs = len(reference_sample)
    f_0 = 0

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2 > 0:
        f_0 = ((sample_size - 1) / sample_size) * f_1 ** 2 / (2 * f_2)
    else:
        f_0 = ((sample_size - 1) / sample_size) * f_1 * (f_1 - 1) / 2
    f_0 = math.ceil(f_0)

    bootstrap_sample = {}
    for n in range(0,sample_size):
        for s,p in enumerate(probabilities):
            if random.random() <= p:
                bootstrap_sample[s]=bootstrap_sample.get(s,0) +1

    #print(bootstrap_sample)
    #print(len(bootstrap_sample), sum(bootstrap_sample.values()))
    return bootstrap_sample

#generate multiple bootstrap sample for a set of sample sizes up to the given sample size:
def generate_bootstrap_sequence_abundance(reference_sample, sample_size, step_size=10):
    s_obs = len(reference_sample)

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)
    f_0 = 0
    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0 = math.ceil(f_0)
    #print(s_obs,f_0)
    probabilities = []
    c = get_c_n_abundance(reference_sample, sample_size)
    factor = factor_abundance(reference_sample, sample_size, c)
   # print(c,factor)

    for x_i in reference_sample.values():
        #print(x_i, sample_size)
    #    print((x_i / sample_size), (1 - factor * (1 - (x_i / sample_size) ** sample_size)))
        adapted_p = (x_i / sample_size) * (1 - factor * ((1 - (x_i / sample_size)) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((1 - c) / f_0) for i in range(0, f_0)]

    #print(probabilities)
    #print(sum(probabilities))

    sample_sequence = []
    sample_steps = []
    bootstrap_sample = {}
    steps = 0
    while steps + step_size < sample_size:
        selected_species = choice([x for x in range(0, s_obs + f_0)], step_size, p=probabilities)
        for s in selected_species:
            bootstrap_sample[s] = bootstrap_sample.get(s, 0) + 1
        steps = steps + step_size
        sample_steps.append(steps)
        sample_sequence.append(copy.deepcopy(bootstrap_sample))
        #print(bootstrap_sample)
        #print(steps,len(bootstrap_sample), sum(bootstrap_sample.values()))

    if sample_size - steps >0:
        selected_species = choice([x for x in range(0, s_obs + f_0)], sample_size-steps, p=probabilities)
        for s in selected_species:
            bootstrap_sample[s] = bootstrap_sample.get(s, 0) + 1
        steps = steps+ (sample_size-steps)
        sample_steps.append(steps)
        sample_sequence.append(copy.deepcopy(bootstrap_sample))
        #print(bootstrap_sample)
        #print(steps, len(bootstrap_sample), sum(bootstrap_sample.values()))

    return sample_sequence, sample_steps


#generate a single bootstrap sample for the given sample size
def generate_bootstrap_sample_abundance(reference_sample, sample_size):
    s_obs = len(reference_sample)

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)
    f_0 = 0
    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0 = math.ceil(f_0)
    #print(s_obs,f_0)
    probabilities = []
    c = get_c_n_abundance(reference_sample, sample_size)
    factor = factor_abundance(reference_sample, sample_size, c)

    for x_i in reference_sample.values():
        adapted_p = (x_i / sample_size) * (1 - factor * ((1 - (x_i / sample_size)) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((1 - c) / f_0) for i in range(0, f_0)]

    #print(probabilities)
    #print(sum(probabilities))

    selected_species = choice([x for x in range(0, s_obs + f_0)], sample_size, p=probabilities)
    bootstrap_sample = {}
    for s in selected_species:
        bootstrap_sample[s] = bootstrap_sample.get(s, 0) + 1

    #print(bootstrap_sample)
    #print(len(bootstrap_sample), sum(bootstrap_sample.values()))
    return bootstrap_sample



def generate_bootstrap_sequence_incidence(reference_sample, sample_size, step_size=10):
    s_obs = len(reference_sample)
    f_0 = 0

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0=math.ceil(f_0)
    #print(s_obs,f_0)

    probabilities = []
    c = get_c_n_incidence(reference_sample, sample_size)
    u = sum(reference_sample.values())

    factor = factor_incidence(reference_sample, sample_size, u, c)

    for y_i in reference_sample.values():
        adapted_p = (y_i / sample_size) * (1 - factor * ((1 - (y_i / sample_size)) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((u/sample_size) * (1 - c) / f_0) for i in range(0, f_0)]
    print("PROBABILITIES")
    print(probabilities)
    print(len(probabilities))
    #print(sum(probabilities))

    sample_sequence = []
    sample_steps = []
    steps = 0
    bootstrap_sample = {}
    for n in range(0,sample_size):
        steps= steps + 1
        for s,p in enumerate(probabilities):
            if random.random() <= p:
                bootstrap_sample[s]=bootstrap_sample.get(s,0) +1
        if steps % step_size == 0:
            sample_sequence.append(copy.deepcopy(bootstrap_sample))
            sample_steps.append(steps)
            #print(bootstrap_sample)
    #print(len(bootstrap_sample), steps,sum(bootstrap_sample.values()),sum(reference_sample.values()))

    if steps % step_size > 0:
        sample_sequence.append(copy.deepcopy(bootstrap_sample))
        sample_steps.append(steps)

        #print(bootstrap_sample)
        #print(len(bootstrap_sample), steps)
    return sample_sequence, sample_steps



def generate_bootstrap_sample_incidence(reference_sample, sample_size):
    s_obs = len(reference_sample)
    f_0 = 0

    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2>0:
        f_0 = ((sample_size-1)/sample_size) * f_1**2 / (2*f_2)
    else:
        f_0 = ((sample_size-1)/sample_size) * f_1 * (f_1-1) / 2
    f_0=math.ceil(f_0)
    #print(s_obs,f_0)

    probabilities = []
    c = get_c_n_incidence(reference_sample, sample_size)
    u = sum(reference_sample.values())

    factor = factor_incidence(reference_sample, sample_size, u, c)

    for y_i in reference_sample.values():
        adapted_p = (y_i / sample_size) * (1 - factor * (1 - (y_i / sample_size) ** sample_size))
        probabilities.append(adapted_p)
    [probabilities.append((u/sample_size) * (1 - c) / f_0) for i in range(0, f_0)]

    #print(probabilities)
    #print(sum(probabilities))

    #selected_species = choice([x for x in range(0, s_obs + f_0)], sample_size, p=probabilities)
    bootstrap_sample = {}
    for n in range(0,sample_size):
        for s,p in enumerate(probabilities):
            if random.random() <= p:
                bootstrap_sample[s]=bootstrap_sample.get(s,0) +1

    #print(bootstrap_sample)
    #print(len(bootstrap_sample), sum(bootstrap_sample.values()))
    return bootstrap_sample


def get_bootstrap_stderr(reference_sample, sample_size, q=0, abundance=True, bootstrap_repetitions= 200):
    estimates_q0 = []
    estimates_q1 = []
    estimates_q2 = []


    probabilities = []
    if abundance:
        probabilities = get_bootstrap_probabilities_abundance(reference_sample, sample_size)
    else:
        probabilities = get_bootstrap_probabilities_incidence(reference_sample, sample_size)

    for i in range(0,bootstrap_repetitions):
        sample = None
        if abundance:
            sample = generate_sample_abundance(reference_sample, sample_size, probabilities)
        else:
            sample = generate_sample_incidence(reference_sample, sample_size, probabilities)

        sample_estimate = None
        sample_estimate_q0 = estimate_species_richness_chao(sample, obs_species_count=sample_size)
        sample_estimate_q1 = estimate_shannon_entropy_abundance(sample, sample_size) if abundance else estimate_shannon_entropy_incidence(sample, sample_size)
        sample_estimate_q2 = estimate_simpson_diversity_abundance(sample, sample_size) if abundance else estimate_simpson_diversity_incidence(sample, sample_size)
        estimates_q0.append(sample_estimate_q0)
        estimates_q1.append(sample_estimate_q1)
        estimates_q2.append(sample_estimate_q2)

    if q==0:
        return statistics.stdev(estimates_q0)
    if q==1:
        return statistics.stdev(estimates_q1)
    if q==2:
        return statistics.stdev(estimates_q2)
    if q==-1:
        return(statistics.stdev(estimates_q0),statistics.stdev(estimates_q1),statistics.stdev(estimates_q2))

def factor_abundance(reference_sample, n, c):
    #print(reference_sample, n, c)
    #print(sum([(x_i / n) * (1 - (x_i / n))**n for x_i in reference_sample.values()]))
    #print()
    if c == 1:
        return 0
    return (1 - c) / (sum([(x_i / n) * (1 - (x_i / n))**n for x_i in reference_sample.values()]))


def factor_incidence(reference_sample, n, u, c):
    if c == 1:
        return 0
    return ((u / n) * (1 - c)) / (sum([(x_i / n) * (1 - (x_i / n))**n for x_i in reference_sample.values()]))


def get_c_n_abundance(reference_sample, n):
    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)

    if f_2 > 0:
        return 1 - (f_1 / n) * (((n - 1) * f_1) / ((n - 1) * f_1 + 2 * f_2))
    else:
        return 1 - (f_1 / n) * (((n - 1) * (f_1 - 1)) / ((n - 1) * (f_1 - 1) + 2))


def get_c_n_incidence(reference_sample, n):
    u = sum(reference_sample.values())
    f_1 = get_singletons(reference_sample)
    f_2 = get_doubletons(reference_sample)
    if f_2 > 0:
        return 1-(f_1/ u) * (((n-1)*f_1)/((n-1)*f_1+2*f_2))
    else:
        return 1-(f_1/ u) * (((n-1)*(f_1-1))/((n-1)*(f_1-1)+2))


import copy
from functools import partial

import pandas as pd
import pm4py
from pm4py.objects.log.obj import EventLog

import species_estimator
import species_retrieval
from plots import plot_rank_abundance


def profile_log(log, name):
    estimators = \
        {
            "1-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=1),
                                                         quantify_all=True),
            "2-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=2),
                                                         quantify_all=True),
            "3-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=3),
                                                         quantify_all=True),
            "4-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=4),
                                                         quantify_all=True),
            "5-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=5),
                                                         quantify_all=True),
            "trace_variants": species_estimator.SpeciesEstimator(species_retrieval.retrieve_species_trace_variant,
                                                                 quantify_all=True),
            "est_act_1" : species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_timed_activity, interval_size=1), quantify_all=True),
            "est_act_5" : species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_timed_activity, interval_size=5), quantify_all=True),
            "est_act_30" : species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_timed_activity, interval_size=30), quantify_all=True),
            "est_act_exp" : species_estimator.SpeciesEstimator(partial(
                species_retrieval.retrieve_timed_activity_exponential), quantify_all=True)
        }
    metrics_stats = {}
    print("Profiling log " + name)
    for est_id, est in estimators.items():
        print(est_id)
        est.apply(log)
        metrics_stats[est_id] = {}

    print("Saving metrics in csv file")
    for est_id, est in estimators.items():
        for metric in species_estimator.METRICS:
            #print(metric)
            metrics_stats[est_id][metric] = getattr(est, metric)[0]
            print(metric + ": " + str(getattr(est, metric)))
        print(est_id)
        plot_rank_abundance(est, str(name) + " - " + str(est_id))
        print()
        print("Total Abundances: " + str(est.total_number_species_abundances))
        print("Total Incidences: " + str(est.total_number_species_incidences))
        print("Degree of Aggregation: " + str(est.degree_spatial_aggregation))
        print("Observations Incidence: " + str(est.number_observations_incidence))
        print("Species Counts: ")
        [print(x, end=" ") for x in est.reference_sample_incidence.values()]
        print()
    df_stats = pd.DataFrame.from_dict(metrics_stats)
    df_stats.to_csv("results/" + str(name) + "_metrics.csv")



log = pm4py.read_xes("../../logs/Sepsis_Cases_-_Event_Log.xes", return_legacy_log_object=True)

log_pre_admission = copy.deepcopy(log)
log_post_admission = copy.deepcopy(log)

log_young = EventLog(attributes=log.attributes, extensions=log.extensions, omni_present=log.omni_present,
                                classifiers=log.classifiers, properties=log.properties)
log_old = EventLog(attributes=log.attributes, extensions=log.extensions, omni_present=log.omni_present,
                                classifiers=log.classifiers, properties=log.properties)


for t in range(0, len(log)):
    #print(log[t])
    if log[t][0]['Age']>=60:
        log_old.append(log[t])
    else:
        log_young.append(log[t])
    for idx, e in enumerate(log[t]):
        if "Admission" in e["concept:name"]:
            if len(log_pre_admission[t][:idx + 1])>0:
                log_pre_admission[t] = log_pre_admission[t][:idx + 1]

            if len(log_post_admission[t][idx + 1:]) > 0:
                log_post_admission[t] = log_post_admission[t][idx + 1:]
            #print([e["concept:name"] for e in log_a[t]])
            #print([e["concept:name"] for e in log_pre_admission[t]])
            #print([e["concept:name"] for e in log_post_admission[t]])
            #print()
            break


profile_log(log_pre_admission, "log_vs_log_eval_sepsis_cases_pre_admission")
profile_log(log_post_admission, "log_vs_log_eval_sepsis_cases_post_admission")
profile_log(log_young, "log_vs_log_eval_sepsis_cases_age_less_60")
profile_log(log_old, "log_vs_log_eval_sepsis_cases_age_geq_60")



from functools import partial

from pandas import DataFrame

import species_retrieval
import species_estimator

from plots import plot_all_stats
from src.simulation import simulate_model
import pm4py
import copy

STEP_SIZE = 10
#plt.style.use('seaborn-v0_8-ticks')


def profile_log(log, name, tick_locations=[20,40,60,80,100], tick_names=[200,400,600,800,1000]):
    estimators = \
        {
            "1-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=1), step_size=10),
            "2-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=2), step_size=10),
            "3-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=3), step_size=10),
            #"4-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=4), step_size=10),
            #"5-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=5), step_size=10),
            #"trace_variants": species_estimator.SpeciesEstimator(species_retrieval.retrieve_species_trace_variant, step_size=10),
        }

    df_stats = {}

    for key, est in estimators.items():
        print("Profiling Log "+str(name)+" "+str(key))
        est.apply(log)
        df_stats[key]={}

    for est_id, est in estimators.items():
        print(name, est_id)
        for metric in species_estimator.METRICS:
            print(metric +": "+str(getattr(est, metric)[-1]))
            df_stats[est_id][metric]=getattr(est, metric)[-1]

        print("Total Abundances: " + str(est.total_number_species_abundances))
        print("Total Incidences: " + str(est.total_number_species_incidences))
        print("Degree of Aggregation: " + str(est.degree_spatial_aggregation))
        print("Observations Incidence: " + str(est.number_observations_incidence))
        print("Species Counts: ")
        [print(x, end=" ") for x in est.reference_sample_incidence.values()]
        print()
        print()
        plot_all_stats([est], "drift_eval_" + name + "_" + est_id, tick_locations, tick_names)

    df = DataFrame.from_dict(df_stats)
    df.to_csv("results/drift_eval_"+name+"_metrics.csv")



# plot functions
# def plot_rank_abundance(estimation, name):
#     plt.rcParams['figure.figsize'] = [8, 3]
#     plt.rcParams['xtick.labelsize'] = 18
#     plt.rcParams['ytick.labelsize'] = 18
#
#     no_species = len(estimation.reference_sample_abundance.values())
#     plt.plot(sorted(list(estimation.reference_sample_abundance.values()), reverse=True))
#     plt.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
#                      sorted(list(estimation.reference_sample_abundance.values()), reverse=True),
#                      [0 for _ in range(no_species)], alpha=0.5)
#     plt.title("Species Occurences (Abundance)", fontsize=26)
#     plt.xlabel("Species Rank", fontsize=22)
#     plt.ylabel("Occurences", fontsize=22)
#     plt.xticks([0, no_species - 1], [0, no_species - 1])
#     plt.yticks([0, max(estimation.reference_sample_abundance.values()) - 1],
#                [0, max(estimation.reference_sample_abundance.values()) - 1])
#
#     plt.tight_layout()
#     plt.savefig("figures/drift_eval_" + name + "_curve_abundance.pdf", format="pdf")
#     plt.show()
#
#     no_species = len(estimation.reference_sample_incidence.values())
#     plt.plot(sorted(list(estimation.reference_sample_incidence.values()), reverse=True))
#     plt.fill_between(np.linspace(0, no_species, no_species, endpoint=False),
#                      sorted(list(estimation.reference_sample_incidence.values()), reverse=True),
#                      [0 for _ in range(no_species)], alpha=0.5)
#     plt.title("Species Occurence (Incidence)", fontsize=26)
#     plt.xlabel("Species Rank", fontsize=22)
#     plt.ylabel("Occurences", fontsize=22)
#     plt.xticks([0, no_species - 1], [0, no_species - 1])
#     plt.yticks([0, max(estimation.reference_sample_incidence.values()) - 1],
#                [0, max(estimation.reference_sample_incidence.values()) - 1])
#
#     plt.tight_layout()
#     plt.savefig("figures/drift_eval_" + name + "_curve_incidence.pdf", format="pdf")
#     plt.show()


# def plot_all_stats(estimation, tick_locations, tick_names, name):
#     labels = estimation.observation_ids
#     no_data_points = len(labels)
#     # no_repetitions = len(estimation)
#     no_observations_abundance = estimation.number_observations_abundance
#     no_observations_incidence = estimation.number_observations_incidence
#
#     obs_stats = estimation.observed_species_count
#     chao2_abundance_stats = estimation.chao2_abundance
#     chao2_incidence_stats = estimation.chao2_incidence
#
#     D1_abundance_sample_stats = estimation.D1_sample_abundance
#     D1_incidence_sample_stats = estimation.D1_sample_incidence
#     D1_abundance_estimated_stats = estimation.D1_estimated_abundance
#     D1_incidence_estimated_stats = estimation.D1_estimated_incidence
#
#     D2_abundance_sample_stats = estimation.D2_sample_abundance
#     D2_incidence_sample_stats = estimation.D2_sample_incidence
#     D2_abundance_estimated_stats = estimation.D2_estimated_abundance
#     D2_incidence_estimated_stats = estimation.D2_estimated_incidence
#
#     coverage_abundance_stats = estimation.coverage_abundance
#     coverage_incidence_stats = estimation.coverage_incidence
#
#     completeness_abundance_stats = estimation.completeness_abundance
#     completeness_incidence_stats = estimation.completeness_incidence
#
#     plt.rcParams['figure.figsize'] = [28, 5]
#     plt.rcParams['xtick.labelsize'] = 18
#     plt.rcParams['ytick.labelsize'] = 18
#
#     # abundance plots
#     # NOTE: we only collect stats after trace level, thus we plot abundance-based stats using incidence observation counts
#     # as number of observations instead of abundance observation counts
#
#     f, (ax1, ax2, ax3), = plt.subplots(nrows=1, ncols=3, sharey='all')
#     y_max = (max(max(obs_stats), max(chao2_abundance_stats), max(chao2_incidence_stats))) + 1
#
#     ax1.plot(obs_stats, label="Observed")
#     ax1.plot(chao2_abundance_stats, label="Estimated")
#     ax1.set_title("q=0 (Abundance)", fontsize=26)
#     ax1.set_xlabel("Sample Size", fontsize=22)
#     ax1.set_ylabel("Hill number", fontsize=22)
#     ax1.set_xticks(tick_locations)
#     ax1.set_xticklabels(tick_names)
#     ax1.tick_params(labelleft=True)
#     ax1.set_ylim(bottom=0, top=y_max)
#     ax1.legend()
#
#     ax1.figure.savefig('TEST.pdf')
#
#     ax2.plot(D1_abundance_sample_stats, label="Observed")
#     ax2.plot(D1_abundance_estimated_stats, label="Estimated")
#     ax2.set_title("q=1 (Abundance)", fontsize=26)
#     ax2.set_xlabel("Sample Size", fontsize=22)
#     ax2.set_ylabel("Hill number", fontsize=22)
#     ax2.set_xticks(tick_locations)
#     ax2.set_xticklabels(tick_names)
#     ax2.tick_params(labelleft=True)
#     ax2.set_ylim(bottom=0, top=y_max)
#     ax2.legend()
#
#     ax3.plot(D2_abundance_sample_stats, label="Observed")
#     ax3.plot(D2_abundance_estimated_stats, label="Estimated")
#     ax3.set_title("q=2 (Abundance)", fontsize=26)
#     ax3.set_xlabel("Sample Size", fontsize=22)
#     ax3.set_ylabel("Hill number", fontsize=22)
#     ax3.set_xticks(tick_locations)
#     ax3.set_xticklabels(tick_names)
#     ax3.tick_params(labelleft=True)
#     ax3.set_ylim(bottom=0, top=y_max)
#     ax3.legend()
#
#     plt.tight_layout()
#     f.savefig("figures/drift_eval_" + name + "_abundance.pdf", format="pdf")
#     plt.show()
#
#     f, (ax4, ax5, ax6) = plt.subplots(nrows=1, ncols=3, sharey='all')
#
#     ax4.plot(obs_stats, label="Observed")
#     ax4.plot(chao2_incidence_stats, label="Estimated")
#     ax4.set_title("q=0 (Incidence)", fontsize=26)
#     ax4.set_xlabel("Sample Size", fontsize=22)
#     ax4.set_ylabel("Hill number", fontsize=22)
#     ax4.set_xticks(tick_locations)
#     ax4.set_xticklabels(tick_names)
#     ax4.tick_params(labelleft=True)
#     ax4.set_ylim(bottom=0, top=y_max)
#     ax4.legend()
#
#     ax5.plot(D1_incidence_sample_stats, label="Observed")
#     ax5.plot(D1_incidence_estimated_stats, label="Estimated")
#     ax5.set_title("q=1 (Incidence)", fontsize=26)
#     ax5.set_xlabel("Sample Size", fontsize=22)
#     ax5.set_ylabel("Hill number", fontsize=22)
#     ax5.set_xticks(tick_locations)
#     ax5.set_xticklabels(tick_names)
#     ax5.tick_params(labelleft=True)
#     ax5.set_ylim(bottom=0, top=y_max)
#     ax5.legend()
#
#     ax6.plot(D2_incidence_sample_stats, label="Observed")
#     ax6.plot(D2_incidence_estimated_stats, label="Estimated")
#     ax6.set_title("q=2 (Incidence)", fontsize=26)
#     ax6.set_xlabel("Sample Size", fontsize=22)
#     ax6.set_ylabel("Hill number", fontsize=22)
#     ax6.set_xticks(tick_locations)
#     ax6.set_xticklabels(tick_names)
#     ax6.tick_params(labelleft=True)
#     ax6.set_ylim(bottom=0, top=y_max)
#     ax6.legend()
#
#     plt.tight_layout()
#     f.savefig("figures/drift_eval_" + name + "_incidence.pdf", format="pdf")
#     plt.show()
#
#     plt.rcParams['figure.figsize'] = [9, 4]
#
#
#     plt.plot(completeness_abundance_stats, label="Completeness")
#     plt.plot(coverage_abundance_stats, label="Coverage")
#     plt.title("Completeness & Coverage", fontsize=26)
#     plt.xlabel("Sample Size", fontsize=22)
#     plt.ylabel("Completeness/Coverage", fontsize=22)
#     plt.xticks(tick_locations, tick_names)
#     plt.ylim(bottom=0)
#     plt.legend()
#
#     plt.tight_layout()
#     f.savefig("figures/drift_eval_" + name + "_coverage_completeness_abundance.pdf", format="pdf")
#     plt.show()
#
#     plt.plot(completeness_incidence_stats, label="Completeness")
#     plt.plot(coverage_incidence_stats, label="Coverage")
#     plt.title("Completeness & Coverage", fontsize=26)
#     plt.xlabel("Sample Size", fontsize=22)
#     plt.ylabel("Completeness/Coverage", fontsize=22)
#     plt.xticks(tick_locations, tick_names)
#     plt.ylim(bottom=0)
#
#     plt.legend()
#
#     plt.tight_layout()
#     f.savefig("figures/drift_eval_" + name + "_coverage_completeness_incidence.pdf", format="pdf")
#     plt.show()


BASE_LENGTH = 1000
MODIFIED_LENGTH = 1000

net, im, fm = pm4py.read_pnml("../../models/base_net.pnml")
#pm4py.view_petri_net(net, im, fm)

base_log = simulate_model(net, im, fm, 1, BASE_LENGTH)

net, im, fm = pm4py.read_pnml("../../models/net_added_e.pnml")
#pm4py.view_petri_net(net, im, fm)

added_e_log = simulate_model(net, im, fm, 1, MODIFIED_LENGTH)

net, im, fm = pm4py.read_pnml("../../models/net_removed_b.pnml")
#pm4py.view_petri_net(net, im, fm)

removed_b_log = simulate_model(net, im, fm, 1, MODIFIED_LENGTH)

net, im, fm = pm4py.read_pnml("../../models/net_equal_prob.pnml")
#pm4py.view_petri_net(net, im, fm)

equal_prob_log = simulate_model(net, im, fm, 1, MODIFIED_LENGTH)

profile_log(base_log[0], "base_model", [0, BASE_LENGTH/STEP_SIZE],[0, BASE_LENGTH])

profile_log(added_e_log[0], "added_activity", [0, MODIFIED_LENGTH/STEP_SIZE],[0, MODIFIED_LENGTH])

profile_log(removed_b_log[0], "removed_activity", [0, MODIFIED_LENGTH/STEP_SIZE],[0, MODIFIED_LENGTH])

profile_log(equal_prob_log[0], "equal_probability", [0, MODIFIED_LENGTH/STEP_SIZE],[0, MODIFIED_LENGTH])



a = copy.deepcopy(base_log[0])
b = copy.deepcopy(added_e_log[0])
for x in b:
    a.append(x)
print(type(b))
print(len(a))
profile_log(a, "base_to_added",[0,100,200], [0,1000,2000])#

a = copy.deepcopy(base_log[0])
b = copy.deepcopy(removed_b_log[0])
for x in b:
    a.append(x)
print(type(b))
print(len(a))
profile_log(a, "base_to_removed",[0,100,200], [0,1000,2000])#

a = copy.deepcopy(base_log[0])
b = copy.deepcopy(equal_prob_log[0])
for x in b:
    a.append(x)
print(type(b))
print(len(a))
profile_log(a, "base_to_equal",[0,100,200], [0,1000,2000])#


from functools import partial

import pandas as pd
import pm4py

import species_estimator
import species_retrieval
from src.simulation import simulate_model


def profile_log(log, name):
    print("Profiling log " + name)

    estimators = \
        {
            "1-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=1),
                                                         quantify_all=True, hill="q0"),
            "2-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=2),
                                                         quantify_all=True, hill="q0"),
            "3-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=3),
                                                         quantify_all=True, hill="q0"),
            "4-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=4),
                                                         quantify_all=True, hill="q0"),
            "5-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=5),
                                                         quantify_all=True, hill="q0"),
            "trace_variants": species_estimator.SpeciesEstimator(species_retrieval.retrieve_species_trace_variant,
                                                                 quantify_all=True, hill="q0")
        }
    metrics_stats = {}
    for est_id, est in estimators.items():
        print(name, est_id)
        est.apply(log)
        metrics_stats[est_id] = {}

    print("Saving metrics in csv file")
    for est_id, est in estimators.items():
        print(name, est_id)
        for metric in species_estimator.METRICS_q1:
            metrics_stats[est_id][metric] = getattr(est, metric)[-1]
            print(metric + ": " + str(getattr(est, metric)))
        #plot_rank_abundance(est, str(name) + " - " + str(est_id))
        print()
        print("Total Abundances: " + str(est.total_number_species_abundances))
        print("Total Incidences: " + str(est.total_number_species_incidences))
        print("Degree of Aggregation: " + str(est.degree_spatial_aggregation))
        print("Observations Incidence: " + str(est.number_observations_incidence))
        #print("Species Counts: ")
        #[print(x, end=" ") for x in est.reference_sample_incidence.values()]
        print()
    df_stats = pd.DataFrame.from_dict(metrics_stats)
    df_stats.to_csv("results/" + str(name) + "_metrics.csv")



log = pm4py.read_xes("../../logs/BPI_Challenge_2018.xes", return_legacy_log_object=True)

print("###1###")
mod, i, f = pm4py.discover_petri_net_alpha(log)
#print(i,f)
pm4py.save_vis_petri_net(mod, i, f, "BPI_2018_ALPHA_pm4py.pdf")
log_d_alpha = simulate_model(mod, i, f, 1, 5000)
#print(log_d_alpha[0])
profile_log(log_d_alpha[0], "log_vs_model_BPI_2018_eval_alpha_pm4py")
#print()

#print("###1.5###")
#mod, i, f = pm4py.discover_petri_net_alpha_plus(log)
#print(i,f)
#pm4py.save_vis_petri_net(mod, i, f, "ALPHA_PLUS_pm4py.pdf")
#log_d_alpha = simulate_model(mod, i, f, 1, 13087)
#profile_log(log_d_alpha[0], "log_vs_model_eval_alpha_plus_pm4py")

print("###2.1###")
mod2, i, f = pm4py.discover_petri_net_inductive(log, noise_threshold=0.2)
print(i,f)
pm4py.save_vis_petri_net(mod2, i, f, "BPI_2018_INDUCTIVE_infrequent_0.2_pm4py.pdf")
log_d_ind = simulate_model(mod2, i, f, 1, 5000)
profile_log(log_d_ind[0], "log_vs_model_eval_BPI_2018_inductive_infrequent_0.2_pm4py")

print("###2.2###")
mod2, i, f = pm4py.discover_petri_net_inductive(log, noise_threshold=0.4)
print(i,f)
pm4py.save_vis_petri_net(mod2, i, f, "BPI_2018_INDUCTIVE_infrequent_0.4_pm4py.pdf")
log_d_ind = simulate_model(mod2, i, f, 1, 5000)
profile_log(log_d_ind[0], "log_vs_model_eval_BPI_2018_inductive_infrequent_0.4_pm4py")


print("###2.3###")
mod2, i, f = pm4py.discover_petri_net_inductive(log, noise_threshold=0.6)
print(i,f)
pm4py.save_vis_petri_net(mod2, i, f, "BPI_2018_INDUCTIVE_infrequent_0.6_pm4py.pdf")
log_d_ind = simulate_model(mod2, i, f, 1, 5000)
profile_log(log_d_ind[0], "log_vs_model_eval_BPI_2018_inductive_infrequent_0.6_pm4py")


print("###2.4###")
mod2, i, f = pm4py.discover_petri_net_inductive(log, noise_threshold=0.8)
print(i,f)
pm4py.save_vis_petri_net(mod2, i, f, "BPI_2018_INDUCTIVE_infrequent_0.8_pm4py.pdf")
log_d_ind = simulate_model(mod2, i, f, 1, 5000)
profile_log(log_d_ind[0], "log_vs_model_eval_BPI_2018_inductive_infrequent_0.8_pm4py")

print("###2.5###")
mod2, i, f = pm4py.discover_petri_net_inductive(log, noise_threshold=0.0)
print(i,f)
pm4py.save_vis_petri_net(mod2, i, f, "BPI_2018_INDUCTIVE_0.0_pm4py.pdf")
log_d_ind = simulate_model(mod2, i, f, 1, 5000)
profile_log(log_d_ind[0], "log_vs_model_eval_BPI_2018_inductive_0.0_pm4py")

print("###3.1###")
mod3, i, f = pm4py.discover_petri_net_ilp(log, alpha=1.0)
print(i,f)
pm4py.save_vis_petri_net(mod3, i, f, "BPI_2018_ILP_1.0_pm4py.pdf")
log_d_ind2 = simulate_model(mod3, i, f, 1, 5000)
profile_log(log_d_ind2[0], "log_vs_model_eval_BPI_2018_ilp_1.0_pm4py")

print("###3.2###")
mod3, i, f = pm4py.discover_petri_net_ilp(log, alpha=0.8)
print(i,f)
pm4py.save_vis_petri_net(mod3, i, f, "BPI_2018_ILP_0.8_pm4py.pdf")
log_d_ind2 = simulate_model(mod3, i, f, 1, 5000)
profile_log(log_d_ind2[0], "log_vs_model_eval_BPI_2018_ilp_0.8_pm4py")

print("###3.2###")
mod3, i, f = pm4py.discover_petri_net_ilp(log, alpha=0.7)
print(i,f)
pm4py.save_vis_petri_net(mod3, i, f, "BPI_2018_ILP_0.7_pm4py.pdf")
log_d_ind2 = simulate_model(mod3, i, f, 1, 5000)
profile_log(log_d_ind2[0], "log_vs_model_eval_BPI_2018_ilp_0.7_pm4py")

print("###3.3###")
mod3, i, f = pm4py.discover_petri_net_ilp(log, alpha=0.6)
print(i,f)
pm4py.save_vis_petri_net(mod3, i, f, "BPI_2018_ILP_0.6_pm4py.pdf")
log_d_ind2 = simulate_model(mod3, i, f, 1, 5000)
profile_log(log_d_ind2[0], "log_vs_model_eval_BPI_2018_ilp_0.6_pm4py")

print("###3.4###")
mod3, i, f = pm4py.discover_petri_net_ilp(log, alpha=0.4)
print(i,f)
pm4py.save_vis_petri_net(mod3, i, f, "BPI_2018_ILP_0.4_pm4py.pdf")
log_d_ind2 = simulate_model(mod3, i, f, 1, 5000)
profile_log(log_d_ind2[0], "log_vs_model_eval_BPI_2018_ilp_0.4_pm4py")

print("###3.5###")
mod3, i, f = pm4py.discover_petri_net_ilp(log, alpha=0.2)
print(i,f)
pm4py.save_vis_petri_net(mod3, i, f, "BPI_2018_ILP_0.2_pm4py.pdf")
log_d_ind2 = simulate_model(mod3, i, f, 1, 5000)
profile_log(log_d_ind2[0], "log_vs_model_eval_BPI_2018_ilp_0.2_pm4py")


# print("###5###")
# net, im, fm = pm4py.read_pnml("models/bpi_2012_alpha.pnml")
# print(im,fm)
# pm4py.save_vis_petri_net(net, im, fm, "ALPHA.pdf")
# log_alpha = simulate_model(net, im, fm, 1, 20000)
# profile_log(log_alpha[0], "log_vs_model_eval_alpha")
#
# print("###6###")
# net, im, fm = pm4py.read_pnml("models/bpi_2012_im.pnml")
# print(im,fm)
# pm4py.save_vis_petri_net(net, im, fm,"INDUCTIVE.pdf")
# log_im = simulate_model(net, im, fm, 1, 20000)
# profile_log(log_im[0], "log_vs_model_eval_inductive")
#
# print("###7###")
# net, im, fm = pm4py.read_pnml("models/bpi_2012_im_infrequent.pnml")
# print(im,fm)
# pm4py.save_vis_petri_net(net, im, fm, "INDUCTIVE_infrequent.pdf")
# log_im_infrequent = simulate_model(net, im, fm, 1, 20000)
# profile_log(log_im_infrequent[0], "log_vs_model_eval_inductive_infrequent")

#print("###8###")
#net, im, fm = pm4py.read_pnml("models/bpi_2012_im_incomplete.pnml")
#pm4py.save_vis_petri_net(net, im, fm, "INDUCTIVE_incomplete.pdf")
#log_im_incomplete = simulate_model(net, im, fm, 1, 20000)
#profile_log(log_im_incomplete[0], "log_vs_model_eval_inductive_incomplete")




from typing import Callable

import pandas as pd
import pm4py
from pandas import DataFrame
from pm4py.objects.log.obj import EventLog, Trace
from tqdm import tqdm

from special4pm.estimation.metrics import get_singletons, get_doubletons, completeness, coverage, \
    sampling_effort_abundance, sampling_effort_incidence, hill_number_asymptotic, entropy_exp, simpson_diversity

# TODO enum for proper key access


class MetricManager(dict):
    # TODO convert to dataclass
    """
    Manages metrics for abundance and incidence models.
    """

    def __init__(self, d0: bool, d1: bool, d2: bool, c0: bool, c1: bool, l_n: list) -> None:
        # reference sample stats
        super().__init__()
        self.reference_sample_abundance = {}
        self.reference_sample_incidence = {}

        self.incidence_current_total_species_count = 0
        self.abundance_current_total_species_count = 0
        self.incidence_sample_size = 0
        self.abundance_sample_size = 0
        self.current_spatial_aggregation = 0

        self["abundance_no_observations"] = [0]
        self["incidence_no_observations"] = [0]
        self["abundance_sum_species_counts"] = [0]
        self["incidence_sum_species_counts"] = [0]
        self["degree_of_aggregation"] = [0]
        self["abundance_singletons"] = [0]
        self["incidence_singletons"] = [0]
        self["abundance_doubletons"] = [0]
        self["incidence_doubletons"] = [0]

        if d0:
            self["abundance_sample_d0"] = [0]
            self["incidence_sample_d0"] = [0]
            self["abundance_estimate_d0"] = [0]
            self["incidence_estimate_d0"] = [0]

        if d1:
            self["abundance_sample_d1"] = [0]
            self["incidence_sample_d1"] = [0]
            self["abundance_estimate_d1"] = [0]
            self["incidence_estimate_d1"] = [0]

        if d2:
            self["abundance_sample_d2"] = [0]
            self["incidence_sample_d2"] = [0]
            self["abundance_estimate_d2"] = [0]
            self["incidence_estimate_d2"] = [0]

        if c0:
            self["abundance_c0"] = [0]
            self["incidence_c0"] = [0]

        if c1:
            self["abundance_c1"] = [0]
            self["incidence_c1"] = [0]

        for l in l_n:
            self["abundance_l_" + str(l)] = [0]
            self["incidence_l_" + str(l)] = [0]


class SpeciesEstimator:
    """
    A class for the estimation of diversity and completeness profiles of trace-based species definitions
    """

    def __init__(self, d0: bool = True, d1: bool = True, d2: bool = True, c0: bool = True,
                 c1: bool = True,
                 l_n: list = [.9, .95, .99], step_size: int | None = None):
        """
        :param d0: flag indicating if D0(=species richness) should be included
        :param d1: flag indicating if D1(=exponential Shannon entropy) should be included
        :param d2: flag indicating if D2(=Simpson diversity index) should be included
        :param c0: flag indicating if C0(=completeness) should be included
        :param c1: flag indicating if C1(=coverage) should be included
        :param l_n: list of desired completeness values for estimation additional sampling effort
        :param step_size: the number of added traces after which the profiles are updated. Use None if
        """
        # TODO add differentiation between abundance and incidence based data
        self.include_abundance = True
        self.include_incidence = True

        self.include_d0 = d0
        self.include_d1 = d1
        self.include_d2 = d2

        self.include_c0 = c0
        self.include_c1 = c1

        self.l_n = l_n

        self.step_size = step_size

        self.metrics = {}
        self.species_retrieval = {}

    def register(self, species_id: str, function: Callable) -> None:
        self.species_retrieval[species_id] = function
        self.metrics[species_id] = MetricManager(self.include_d0, self.include_d1, self.include_d2, self.include_c0,
                                                 self.include_c1, self.l_n)

    def apply(self, data: pd.DataFrame | EventLog | Trace) -> None:
        """
        add all observations of an event log and update diversity and completeness profiles once afterward.
        If parameter step_size is set to an int, profiles are additionally updated along the way according to
        the step size
        :param data: the event log containing the trace observations
        """
        if isinstance(data, pd.DataFrame):
            return self.apply(pm4py.convert_to_event_log(data))

        #todo find out why this is notably faster than self.apply(tr) for tr in Log
        elif isinstance(data, EventLog):
            for species_id in self.species_retrieval.keys():
                for tr in tqdm(data, "Profiling Log for " + species_id):
                    self.add_observation(tr, species_id)
                    # if step size is set, update metrics after <step_size> many traces
                    if self.step_size is None:
                        continue
                    elif self.metrics[species_id].incidence_sample_size % self.step_size == 0:
                        self.update_metrics(species_id)
                self.update_metrics(species_id)

        elif isinstance(data, Trace):
            for species_id in self.species_retrieval.keys():
                self.add_observation(data, species_id)
                # if step size is set, update metrics after <step_size> many traces
                if self.step_size is None:
                    continue
                elif self.metrics[species_id].incidence_sample_size % self.step_size == 0:
                    self.update_metrics(species_id)

        else:
            raise RuntimeError('Cannot apply data of type ' + str(type(data)))



    def add_observation(self, observation: Trace, species_id: str) -> None:
        """
        adds a single observation
        :param species_id: the species definition for which observation shall be added
        :param observation: the trace observation
        """
        # retrieve species from current observation
        species_abundance = self.species_retrieval[species_id](observation)
        species_incidence = set(species_abundance)
        self.metrics[species_id].trace_retrieved_species_abundance = species_abundance
        self.metrics[species_id].trace_retrieved_species_incidence = species_incidence

        # update species abundances/incidences
        for s in species_abundance:
            self.metrics[species_id].reference_sample_abundance[s] = self.metrics[
                                                                         species_id].reference_sample_abundance.get(s,
                                                                                                                    0) + 1

        for s in species_incidence:
            self.metrics[species_id].reference_sample_incidence[s] = self.metrics[
                                                                         species_id].reference_sample_incidence.get(s,
                                                                                                                    0) + 1

        # update current number of observation for each model
        self.metrics[species_id].abundance_sample_size = self.metrics[species_id].abundance_sample_size + len(
            species_abundance)
        self.metrics[species_id].incidence_sample_size = self.metrics[species_id].incidence_sample_size + 1

        # update current sum of all observed species for each model
        self.metrics[species_id].abundance_current_total_species_count = \
            self.metrics[species_id].abundance_current_total_species_count + len(species_abundance)
        self.metrics[species_id].incidence_current_total_species_count = \
            self.metrics[species_id].incidence_current_total_species_count + len(
                species_incidence)

        #update current degree of spatial aggregation
        self.metrics[species_id].current_spatial_aggregation = 1 - (
                self.metrics[species_id].incidence_current_total_species_count / self.metrics[
            species_id].abundance_current_total_species_count)

    def update_metrics(self, species_id: str) -> None:
        """
        updates the diversity and completeness profiles based on the current observations
        """
        # update number of observations so far
        self.metrics[species_id]["abundance_no_observations"].append(self.metrics[species_id].abundance_sample_size)
        self.metrics[species_id]["incidence_no_observations"].append(self.metrics[species_id].incidence_sample_size)

        #update number of species seen so far
        self.metrics[species_id]["abundance_sum_species_counts"].append(
            self.metrics[species_id].abundance_current_total_species_count)
        self.metrics[species_id]["incidence_sum_species_counts"].append(
            self.metrics[species_id].incidence_current_total_species_count)

        #update degree of spatial aggregation
        self.metrics[species_id]["degree_of_aggregation"].append(self.metrics[species_id].current_spatial_aggregation)

        #update singleton and doubleton counts
        self.metrics[species_id]["abundance_singletons"].append(
            get_singletons(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_singletons"].append(
            get_singletons(self.metrics[species_id].reference_sample_incidence))

        self.metrics[species_id]["abundance_doubletons"].append(
            get_doubletons(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_doubletons"].append(
            get_doubletons(self.metrics[species_id].reference_sample_incidence))

        #update diversity profile
        if self.include_d0:
            self.__update_d0(species_id)
        if self.include_d1:
            self.__update_d1(species_id)
        if self.include_d2:
            self.__update_d2(species_id)

        #update completeness profile
        if self.include_c0:
            self.__update_c0(species_id)
        if self.include_c1:
            self.__update_c1(species_id)

        #update estimated sampling effort for target completeness
        for l in self.l_n:
            self.__update_l(l, species_id)

    def __update_d0(self, species_id: str) -> None:
        """
        updates D0 (=species richness) based on the current observations
        """
        #update sample metrics
        self.metrics[species_id]["abundance_sample_d0"].append(len(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_sample_d0"].append(len(self.metrics[species_id].reference_sample_incidence))

        #update estimated metrics
        self.metrics[species_id]["abundance_estimate_d0"].append(
            hill_number_asymptotic(0, self.metrics[species_id].reference_sample_abundance,
                                   self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_estimate_d0"].append(
            hill_number_asymptotic(0, self.metrics[species_id].reference_sample_incidence,
                                   self.metrics[species_id].incidence_sample_size, abundance=False))

    def __update_d1(self, species_id: str) -> None:
        """
        updates D1 (=exponential of Shannon entropy) based on the current observations
        """
        #update sample metrics
        self.metrics[species_id]["abundance_sample_d1"].append(
            entropy_exp(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_sample_d1"].append(
            entropy_exp(self.metrics[species_id].reference_sample_incidence))

        #update estimated metrics
        self.metrics[species_id]["abundance_estimate_d1"].append(
            hill_number_asymptotic(1, self.metrics[species_id].reference_sample_abundance,
                                   self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_estimate_d1"].append(
            hill_number_asymptotic(1, self.metrics[species_id].reference_sample_incidence,
                                   self.metrics[species_id].incidence_sample_size, abundance=False))

    def __update_d2(self, species_id: str) -> None:
        """
        updates D2 (=Simpson Diversity Index) based on the current observations
        """
        #update sample metrics
        self.metrics[species_id]["abundance_sample_d2"].append(
            simpson_diversity(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_sample_d2"].append(
            simpson_diversity(self.metrics[species_id].reference_sample_incidence))

        #update estimated metrics
        self.metrics[species_id]["abundance_estimate_d2"].append(
            hill_number_asymptotic(2, self.metrics[species_id].reference_sample_abundance,
                                   self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_estimate_d2"].append(
            hill_number_asymptotic(2, self.metrics[species_id].reference_sample_incidence,
                                   self.metrics[species_id].incidence_sample_size, abundance=False))

    def __update_c0(self, species_id: str) -> None:
        """
        updates C0 (=completeness) based on the current observations
        """
        self.metrics[species_id]["abundance_c0"].append(
            completeness(self.metrics[species_id].reference_sample_abundance))
        self.metrics[species_id]["incidence_c0"].append(
            completeness(self.metrics[species_id].reference_sample_incidence))

    def __update_c1(self, species_id: str) -> None:
        """
        updates C1 (=coverage) based on the current observations
        """
        self.metrics[species_id]["abundance_c1"].append(
            coverage(self.metrics[species_id].reference_sample_abundance,
                     self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_c1"].append(
            coverage(self.metrics[species_id].reference_sample_incidence,
                     self.metrics[species_id].incidence_sample_size))

    def __update_l(self, g: float, species_id: str) -> None:
        """
        updates l_g (=expected number additional observations for reaching completeness g) based on the current
        observations
        :param g: desired  completeness
        """
        self.metrics[species_id]["abundance_l_" + str(g)].append(
            sampling_effort_abundance(g, self.metrics[species_id].reference_sample_abundance,
                                      self.metrics[species_id].abundance_sample_size))
        self.metrics[species_id]["incidence_l_" + str(g)].append(
            sampling_effort_incidence(g, self.metrics[species_id].reference_sample_incidence,
                                      self.metrics[species_id].incidence_sample_size))

    def print_metrics(self) -> None:
        """
        prints the Diversity and Completeness Profile of the current observations
        """
        for species_id in self.species_retrieval:
            print("### " + species_id + " ###")
            print("### SAMPLE STATS ###")
            print("Abundance")
            print("%-30s %s" % ("     No Observations:", str(self.metrics[species_id]["abundance_no_observations"])))
            print("%-30s %s" % (
                "     Total Species Count:", str(self.metrics[species_id]["abundance_sum_species_counts"])))
            print("%-30s %s" % ("     Singletons:", str(self.metrics[species_id]["abundance_singletons"])))
            print("%-30s %s" % ("     Doubletons:", str(self.metrics[species_id]["abundance_doubletons"])))
            print("Incidence")
            print("%-30s %s" % ("     No Observations:", str(self.metrics[species_id]["incidence_no_observations"])))
            print("%-30s %s" % (
                "     Total Species Count:", str(self.metrics[species_id]["incidence_sum_species_counts"])))
            print("%-30s %s" % ("     Singletons:", str(self.metrics[species_id]["incidence_singletons"])))
            print("%-30s %s" % ("     Doubletons:", str(self.metrics[species_id]["incidence_doubletons"])))
            print("%-30s %s" % ("Degree of Aggregation:", str(self.metrics[species_id]["degree_of_aggregation"])))
            print()
            print("### DIVERSITY AND COMPLETENESS PROFILE ###")
            print("Abundance")
            if self.include_d0:
                print("%-30s %s" % ("     D0 - sample:", (self.metrics[species_id]["abundance_sample_d0"])))
                print("%-30s %s" % ("     D0 - estimate:", str(self.metrics[species_id]["abundance_estimate_d0"])))
            if self.include_d1:
                print("%-30s %s" % ("     D1 - sample:", str(self.metrics[species_id]["abundance_sample_d1"])))
                print("%-30s %s" % ("     D1 - estimate:", str(self.metrics[species_id]["abundance_estimate_d1"])))
            if self.include_d2:
                print("%-30s %s" % ("     D2 - sample:", str(self.metrics[species_id]["abundance_sample_d2"])))
                print("%-30s %s" % ("     D2 - estimate:", str(self.metrics[species_id]["abundance_estimate_d2"])))
            if self.include_c0:
                print("%-30s %s" % ("     C0:", str(self.metrics[species_id]["abundance_c0"])))
            if self.include_c1:
                print("%-30s %s" % ("     C1:", str(self.metrics[species_id]["abundance_c1"])))
            for l in self.l_n:
                print("%-30s %s" % ("     l_" + str(l) + ":", str(self.metrics[species_id]["abundance_l_" + str(l)])))
            print("Incidence")
            if self.include_d0:
                print("%-30s %s" % ("     D0 - sample:", (self.metrics[species_id]["incidence_sample_d0"])))
                print("%-30s %s" % ("     D0 - estimate:", str(self.metrics[species_id]["incidence_estimate_d0"])))
            if self.include_d1:
                print("%-30s %s" % ("     D1 - sample:", str(self.metrics[species_id]["incidence_sample_d1"])))
                print("%-30s %s" % ("     D1 - estimate:", str(self.metrics[species_id]["incidence_estimate_d1"])))
            if self.include_d2:
                print("%-30s %s" % ("     D2 - sample:", str(self.metrics[species_id]["incidence_sample_d2"])))
                print("%-30s %s" % ("     D2 - estimate:", str(self.metrics[species_id]["incidence_estimate_d2"])))
            if self.include_c0:
                print("%-30s %s" % ("     C0:", str(self.metrics[species_id]["incidence_c0"])))
            if self.include_c1:
                print("%-30s %s" % ("     C1:", str(self.metrics[species_id]["incidence_c1"])))
            for l in self.l_n:
                print("%-30s %s" % ("     l_" + str(l) + ":", str(self.metrics[species_id]["incidence_l_" + str(l)])))
            print()

    def to_dataFrame(self) -> DataFrame:
        """
        returns the diversity and completeness profile of the current observations as a data frame
        :returns: a data frame view of the Diversity and Completeness Profile
        """
        return pd.DataFrame([[i, j, ix, v]
                             for i in self.metrics.keys()
                             for j in self.metrics[i].keys()
                             for ix, v in enumerate(self.metrics[i][j])
                             ], columns=["species", "metric", "observation", "value"]
                            )



import species_estimator
import species_retrieval
from functools import partial
import pm4py
import pandas as pd

from src.visualizations.plots import plot_rank_abundance


#plt.style.use('seaborn-v0_8-white')
#plt.style.use('seaborn-v0_8-ticks')

def profile_log(log_path, name):
    log = pm4py.read_xes(log_path, return_legacy_log_object=True)
    print("##### "+name+" #####")
    estimators = \
        {
             "1-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=1),
                                                          quantify_all=True),
             "2-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=2),
                                                          quantify_all=True),
             "3-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=3),
                                                          quantify_all=True),
             "4-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=4),
                                                          quantify_all=True),
             "5-gram": species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=5),
                                                          quantify_all=True),
             "trace_variants": species_estimator.SpeciesEstimator(species_retrieval.retrieve_species_trace_variant,
                                                                  quantify_all=True),
            "est_act_1" : species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_timed_activity, interval_size=1), quantify_all=True),
            "est_act_5" : species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_timed_activity, interval_size=5), quantify_all=True),
            "est_act_30" : species_estimator.SpeciesEstimator(partial(species_retrieval.retrieve_timed_activity, interval_size=30), quantify_all=True),
            "est_act_exp" : species_estimator.SpeciesEstimator(partial(
                species_retrieval.retrieve_timed_activity_exponential), quantify_all=True)
        }

    metrics_stats = {}
    print("Profiling log")
    for est_id,est in estimators.items():
        print(name,est_id)
        est.apply(log)
        metrics_stats[est_id]={}

    print("Saving metrics in csv file")

    for est_id, est in estimators.items():
        print()
        print(name,est_id)
        for metric in species_estimator.METRICS:
            metrics_stats[est_id][metric] = getattr(est, metric)[-1]
            print(metric +": "+str(getattr(est, metric)))

        plot_rank_abundance(est, str(name) + " - " + str(est_id))
        print("Total Abundances: " + str(est.total_number_species_abundances))
        print("Total Incidences: " + str(est.total_number_species_incidences))
        print("Degree of Aggregation: " + str(est.degree_spatial_aggregation))
        print("Observations Incidence: " + str(est.number_observations_incidence))
        print("Species Counts: ")
        [print(x, end=" ") for x in est.reference_sample_incidence.values()]
        print()
    df_stats = pd.DataFrame.from_dict(metrics_stats)
    df_stats.to_csv("results/" + str(name) + "_metrics.csv")

#TODO rarefaction based on bootstrapping! Becomes prohibitively slow for q>=1
    # print("Building Rarefaction and Extrapolation Curves")
    # for est_id, est in estimators.items():
    #     q0, q1, q2 = rarefy_extrapolate_all(est, abundance_data=False, data_points=20)
    #     print(q0[0])
    #     print(q0[1])
    #
    #     plt.rcParams['figure.figsize'] = [6, 5]
    #     plt.rcParams['xtick.labelsize'] = 14
    #     plt.rcParams['ytick.labelsize'] = 14
    #
    #     plt.plot(q0[1][:22], q0[0][:22], label="q=0", color='#F8766D')
    #     plt.plot(q0[1][21:], q0[0][21:], linestyle="--", color='#F8766D')
    #
    #     plt.plot(q1[1][:22], q1[0][:22], label="q=1", color='#00BA38')
    #     plt.plot(q1[1][21:], q1[0][21:], linestyle="--", color='#00BA38')
    #
    #     plt.plot(q2[1][:22], q2[0][:22], label="q=2", color='#619CFF')
    #     plt.plot(q2[1][21:], q2[0][21:], linestyle="--", color='#619CFF')
    #
    #     plt.title(name, fontsize=22)
    #     plt.xlabel("Sample Size", fontsize=18)
    #     plt.ylabel("Diversity", fontsize=18)
    #     plt.xticks([0, est.number_observations_incidence, 2 * est.number_observations_incidence],[0, est.number_observations_incidence, 2 * est.number_observations_incidence])
    #     plt.legend()
    #     plt.tight_layout()
    #     plt.savefig("figures/real_eval_curves_" + name + "_" + est_id +  "_.pdf", format="pdf")
    #
    #     plt.show()
    #save all metrics to csv


profile_log("../../logs/BPI_Challenge_2012.xes", "real_eval_BPI 2012")

profile_log("../../logs/Road_Traffic_Fines_Management_Process.xes", "real_eval_Road Traffic Fines")

profile_log("../../logs/Sepsis_Cases_-_Event_Log.xes", "real_eval_Sepsis Cases")

profile_log("../../logs/BPI_Challenge_2018.xes", "real_eval_BPI 2018")

profile_log("../../logs/BPI_Challenge_2019.xes", "real_eval_BPI 2019")


# ---EXTENSION---#
# TODO bootstrap rarefaction and extrapolation!
# TODO measure generalization of discovered process models
import statistics

import pandas as pd

import src.estimation.species_estimator
import src.estimation.species_retrieval
from src.estimation import species_estimator, species_retrieval
from src.simulation.simulation import simulate_model
from src.visualizations.plots import plot_rank_abundance, plot_all_stats
from src.simulation import simulation
from src.estimation.species_estimator import SpeciesEstimator
from functools import partial
from tqdm import tqdm

import pm4py

ESTIMATORS = ["1-gram", "2-gram", "3-gram", "4-gram", "5-gram", "trace_variants"]

def summarize_metrics_reference_samples(estimations):
    metrics_stats = {}
    for metric in species_estimator.METRICS:
        metrics_reference_sample = sorted([getattr(est,metric)[-1] for est in estimations])
        metrics_stats[metric]=[
            min(metrics_reference_sample),
            metrics_reference_sample[2],
            statistics.mean(metrics_reference_sample),
            statistics.stdev(metrics_reference_sample),
            statistics.median(metrics_reference_sample),
            metrics_reference_sample[-2],
            max(metrics_reference_sample)
        ]
        print(metric + ": " + str(statistics.median(metrics_reference_sample)))
    return pd.DataFrame.from_dict(metrics_stats, orient='index', columns=["min","lower_ci","mean","stddev","median","upper_ci","max"])


def profile_logs(logs):
    #TODO make this a dictionary
    estimations_per_species = [[], [], [], [], [], []]
    for log in tqdm(logs, desc='Evaluating logs'):
        results = profile_log(log)
        for estimator, results in zip(results, estimations_per_species):
            results.append(estimator)
    return estimations_per_species


def profile_log(log):
    est_1gram = SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=1), step_size=20)
    est_2gram = SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=2), step_size=20)
    est_3gram = SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=3), step_size=20)
    est_4gram = SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=4), step_size=20)
    est_5gram = SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=5), step_size=20)
    est_tv = SpeciesEstimator(species_retrieval.retrieve_species_trace_variant, step_size=20)


    estimators = [est_1gram, est_2gram, est_3gram, est_4gram, est_5gram, est_tv]

    # iterate over traces in log
    for est in estimators:
        est.apply(log)
    return estimators


def evaluate_model(path, name, repetitions, log_size):
    net, im, fm = pm4py.read_pnml(path)
    print(name)
    simulated_logs = simulate_model(net, im, fm, repetitions, log_size)
    results_per_species = profile_logs(simulated_logs)

    for result, estimator_name in zip(results_per_species, ESTIMATORS):
        print(name + " - " + estimator_name)
        #TODO summary statistics post-hoc to the csv as df.describe() and df.info()
        summary_df = summarize_metrics_reference_samples(result)
        summary_df.to_csv("results/baseline_eval_"+name+"_"+estimator_name+"_summary.csv")
        plot_rank_abundance(result[0], name + "_" + estimator_name)
        print()
        print("Total Abundances: " + str(result[0].total_number_species_abundances))
        print("Total Incidences: " + str(result[0].total_number_species_incidences))
        print("Degree of Aggregation: " + str(result[0].degree_spatial_aggregation))
        print("Observations Incidence: " + str(result[0].number_observations_incidence))
        print("Species Counts: ")
        [print(x, end=" ") for x in result[0].reference_sample_incidence.values()]
        print()

        plot_all_stats(result, name + "_" + estimator_name, [0, 50, 100], [0,1000,2000])


#evaluate_model("./models/model_1.pnml", "baseline_eval_model_1", 100, 2000)
#evaluate_model("./models/model_2.pnml", "baseline_eval_model_2", 100, 2000)
#evaluate_model("./models/model_4.pnml", "baseline_eval_model_4", 100, 1000)
#evaluate_model("./models/model_5.pnml", "baseline_eval_model_5", 100, 1000)
#evaluate_model("./models/model_6.pnml", "baseline_eval_model_6", 100, 1000)

#Used in Evaluation
evaluate_model("../../models/model_3.pnml", "baseline_eval_model_3", 100, 2000)
#evaluate_model("./models/model_7.pnml", "baseline_eval_model_7_LARGE", 100, 5000)
#evaluate_model("./models/morel_8.pnml", "baseline_eval_model_8_LARGE", 100, 5000)



import copy
import math

import species_estimator
import species_retrieval
from functools import partial

from plots import plot_rank_abundances_multiple_estimators
from species_estimator import SpeciesEstimator
from src.simulation import simulate_model
import pm4py
import random

#plt.style.use('seaborn-v0_8-ticks')


def profile_logs(log_dict):
    estimations = {"1-gram": [], "2-gram": [], "3-gram": [], "4-gram": [], "5-gram": [], "trace_variants": []}
    metrics = species_estimator.METRICS

    for log_id, log in log_dict.items():
        estimators = \
            {
                "1-gram": SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=1),
                                           quantify_all=True),
                "2-gram": SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=2),
                                           quantify_all=True),
                "3-gram": SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=3),
                                           quantify_all=True),
                "4-gram": SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=4),
                                           quantify_all=True),
                "5-gram": SpeciesEstimator(partial(species_retrieval.retrieve_species_n_gram, n=5),
                                           quantify_all=True),
                "trace_variants": SpeciesEstimator(species_retrieval.retrieve_species_trace_variant, quantify_all=True)
            }

        for key, est in estimators.items():
            print("Profiling log "+log_id+" "+key)
            est.apply(log)
            estimations[key].append(est)

    for est_id, estimation in estimations.items():
        print(est_id)
        for log_id, est in zip(log_dict.keys(), estimation):
            print(log_id)
            for metric in metrics:
                print(metric +": "+str(getattr(est, metric)))

            print("Total Abundances: " + str(est.total_number_species_abundances))
            print("Total Incidences: " + str(est.total_number_species_incidences))
            print("Degree of Aggregation: " + str(est.degree_spatial_aggregation))
            print("Observations Incidence: " + str(est.number_observations_incidence))
            print("Species Counts: ")
            [print(x, end=" ") for x in est.reference_sample_incidence.values()]
            print()

        plot_rank_abundances_multiple_estimators(log_dict.keys(), estimation, est_id)






net, im, fm = pm4py.read_pnml("../../models/base_net.pnml")
#pm4py.view_petri_net(net, im, fm)

simulated_log = simulate_model(net, im, fm, 1, 1000)

events = []
for i,tr in enumerate(simulated_log[0]):
    for j,ev in enumerate(tr):
        events.append((i,j))
print("Total Number of events: "+ str(len(events)),"1%: "+str(len(events)/100),"0.1%: "+str(len(events)/1000),"0.01%: "+str(len(events)/10000))

#log consisting of 0.01% noise
log_a = copy.deepcopy(simulated_log[0])
evs = random.sample(events,math.ceil(len(events)/10000))
for ev in evs:
    log_a[ev[0]][ev[1]]["concept:name"] = str(random.getrandbits(128))

#log consisting of 0.1% noise
log_b = copy.deepcopy(simulated_log[0])
evs = random.sample(events,math.ceil(len(events)/1000))
for ev in evs:
    log_b[ev[0]][ev[1]]["concept:name"] = str(random.getrandbits(128))

#log consisting of 1% noise
log_c = copy.deepcopy(simulated_log[0])
evs = random.sample(events,math.ceil(len(events)/100))
for ev in evs:
    log_c[ev[0]][ev[1]]["concept:name"] = str(random.getrandbits(128))

profile_logs({"Original":simulated_log[0],"0.01%":log_a,"0.1%":log_b,"1%":log_c})